import gleam/bool
import gleam/float
import gleam/int
import gleam/list
import gleam/option.{type Option, None, Some}
import gleam/result
import gleam/string

pub type ParseError {
  UnexpectedEndOfFile
  UnexpectedTokenCharacter(got: String, expected: String)
  UnexpectedNameCharacter(got: String)
  UnexpectedNumberCharacter(got: String)
  UnexpectedTrailingString(got: String)
  UnterminatedString(got: String)
  InvalidNumber(got: String)
}

pub type SExpr {
  Token(name: String, attributes: List(SExpr))
  String(String)
  Int(Int)
  Float(Float)
  Name(String)
}

pub type Parsed(a) =
  Result(#(a, String), ParseError)

pub fn sexpr_to_pretty_string(sexpr: SExpr) -> String {
  do_sexpr_to_pretty_string(sexpr, "")
}

fn do_sexpr_to_pretty_string(sexpr: SExpr, pad: String) -> String {
  pad
  <> case sexpr {
    Token(name:, attributes:) ->
      name
      <> " :: Token\n"
      <> attributes
      |> list.map(do_sexpr_to_pretty_string(_, pad <> "  "))
      |> string.join("\n")
    String(value) -> "\"" <> value <> "\" :: String"
    Int(value) -> int.to_string(value) <> " :: Int"
    Float(value) -> float.to_string(value) <> " :: Float"
    Name(value) -> value <> " :: Name"
  }
}

pub fn run(source: String) -> Result(SExpr, ParseError) {
  let source = string.trim(source)
  use #(token, rest) <- result.try(attribute(source))
  case string.trim(rest) {
    "" -> Ok(token)
    rest -> Error(UnexpectedTrailingString(rest))
  }
}

pub fn token(source: String) -> Parsed(SExpr) {
  use #(name, rest) <- result.try(name(source))
  use #(attributes, rest) <- result.try(attributes(rest))
  Ok(#(Token(name:, attributes:), rest))
}

fn name_char(source: String) -> Parsed(String) {
  case source {
    "" -> Error(UnexpectedEndOfFile)
    "a" <> rest -> Ok(#("a", rest))
    "b" <> rest -> Ok(#("b", rest))
    "c" <> rest -> Ok(#("c", rest))
    "d" <> rest -> Ok(#("d", rest))
    "e" <> rest -> Ok(#("e", rest))
    "f" <> rest -> Ok(#("f", rest))
    "g" <> rest -> Ok(#("g", rest))
    "h" <> rest -> Ok(#("h", rest))
    "i" <> rest -> Ok(#("i", rest))
    "j" <> rest -> Ok(#("j", rest))
    "k" <> rest -> Ok(#("k", rest))
    "l" <> rest -> Ok(#("l", rest))
    "m" <> rest -> Ok(#("m", rest))
    "n" <> rest -> Ok(#("n", rest))
    "o" <> rest -> Ok(#("o", rest))
    "p" <> rest -> Ok(#("p", rest))
    "q" <> rest -> Ok(#("q", rest))
    "r" <> rest -> Ok(#("r", rest))
    "s" <> rest -> Ok(#("s", rest))
    "t" <> rest -> Ok(#("t", rest))
    "u" <> rest -> Ok(#("u", rest))
    "v" <> rest -> Ok(#("v", rest))
    "w" <> rest -> Ok(#("w", rest))
    "x" <> rest -> Ok(#("x", rest))
    "y" <> rest -> Ok(#("y", rest))
    "z" <> rest -> Ok(#("z", rest))
    "_" <> rest -> Ok(#("_", rest))
    _ ->
      Error(UnexpectedNameCharacter(string.first(source) |> result.unwrap("")))
  }
}

pub fn name(source: String) -> Parsed(String) {
  use #(char, rest) <- result.try(name_char(source))
  do_name(rest, char)
}

fn do_name(source: String, result: String) -> Parsed(String) {
  case name_char(source) {
    Ok(#(char, rest)) -> do_name(rest, result <> char)
    Error(_) -> Ok(#(result, source))
  }
}

pub fn attributes(source: String) -> Parsed(List(SExpr)) {
  do_attributes(source, [])
}

fn do_attributes(source: String, attributes: List(SExpr)) -> Parsed(List(SExpr)) {
  case string.trim_start(source) {
    "" -> Error(UnexpectedEndOfFile)
    ")" <> rest -> Ok(#(list.reverse(attributes), rest))
    source -> {
      use #(attribute, rest) <- result.try(attribute(source))
      do_attributes(rest, [attribute, ..attributes])
    }
  }
}

pub fn attribute(source: String) -> Parsed(SExpr) {
  case source {
    "" -> Error(UnexpectedEndOfFile)
    "(" <> rest -> {
      use #(token, rest) <- result.try(token(rest))
      Ok(#(token, rest))
    }
    "\"" <> rest -> {
      use #(str, rest) <- result.try(string(rest))
      Ok(#(String(str), rest))
    }
    source -> {
      use <- option.lazy_unwrap(try_number(source))
      use #(name, rest) <- result.try(name(source))
      Ok(#(Name(name), rest))
    }
  }
}

pub fn string(source: String) -> Parsed(String) {
  use <- bool.guard(source == "", Error(UnexpectedEndOfFile))
  do_string(source, "")
}

fn do_string(source: String, result: String) -> Parsed(String) {
  case string.split_once(source, "\"") {
    Ok(#(start, rest)) ->
      case string.last(start) {
        Ok("\\") -> do_string(rest, result <> "\"" <> start)
        _ ->
          Ok(#(
            result <> start,
            // |> string.replace("\\n", "\n")
            // |> string.replace("\\r", "\r")
            // |> string.replace("\\t", "\t")
            // |> string.replace("\\f", "\f")
            rest,
          ))
      }
    Error(Nil) -> Error(UnterminatedString(source))
  }
}

fn number_char(source: String) -> Parsed(String) {
  case source {
    "" -> Error(UnexpectedEndOfFile)
    "." <> rest -> Ok(#(".", rest))
    "-" <> rest -> Ok(#("-", rest))
    "0" <> rest -> Ok(#("0", rest))
    "1" <> rest -> Ok(#("1", rest))
    "2" <> rest -> Ok(#("2", rest))
    "3" <> rest -> Ok(#("3", rest))
    "4" <> rest -> Ok(#("4", rest))
    "5" <> rest -> Ok(#("5", rest))
    "6" <> rest -> Ok(#("6", rest))
    "7" <> rest -> Ok(#("7", rest))
    "8" <> rest -> Ok(#("8", rest))
    "9" <> rest -> Ok(#("9", rest))
    _ ->
      Error(UnexpectedNumberCharacter(string.first(source) |> result.unwrap("")))
  }
}

pub fn try_number(source: String) -> Option(Parsed(SExpr)) {
  case number_char(source) {
    Ok(#(char, rest)) -> Some(do_number(rest, char))
    Error(_) -> None
  }
}

fn do_number(source: String, result: String) -> Parsed(SExpr) {
  case number_char(source) {
    Ok(#(char, rest)) -> do_number(rest, result <> char)
    Error(_) -> {
      case int.parse(result) {
        Ok(n) -> Ok(#(Int(n), source))
        Error(Nil) -> {
          let result = case result {
            "-." <> rest -> "-0." <> rest
            "." <> rest -> "0." <> rest
            result -> result
          }
          case float.parse(result) {
            Ok(n) -> Ok(#(Float(n), source))
            Error(Nil) -> Error(InvalidNumber(result))
          }
        }
      }
    }
  }
}
